@@ -18,15 +18,18 @@ L{Artist}, L{Release}, and L{Track}.
 
 @author: Matthias Friedrich <matt@mafr.de>
 """
-from sets import Set
+try:
+	set
+except NameError:
+	from sets import Set as set
 
-__revision__ = '$Id: model.py 9724 2008-03-09 08:34:11Z matt $'
+__revision__ = '$Id: model.py 12708 2010-03-15 17:45:25Z matt $'
 
 __all__ = [
 	'VARIOUS_ARTISTS_ID', 'NS_MMD_1', 'NS_REL_1', 'NS_EXT_1', 
-	'Entity', 'Artist', 'Release', 'Track', 'User', 
-	'Relation', 'Disc', 'ReleaseEvent', 'Label', 'Tag',
-	'AbstractAlias', 'ArtistAlias', 'LabelAlias', 
+	'Entity', 'Artist', 'Release', 'Track', 'User', 'ReleaseGroup',
+	'Relation', 'Disc', 'ReleaseEvent', 'Label', 'Tag', 'Rating',
+	'AbstractAlias', 'ArtistAlias', 'LabelAlias',
 ]
 
 
@@ -72,6 +75,7 @@ class Entity(object):
 		self._id = id_
 		self._relations = { }
 		self._tags = { }
+		self._rating = Rating()
 
 	def getId(self):
 		"""Returns a MusicBrainz ID.
@@ -146,10 +150,10 @@ class Entity(object):
 		# Now filer for attribute type.
 		#
 		tmp = []
-		required = Set(iter(requiredAttributes))
+		required = set(iter(requiredAttributes))
 
 		for r in allRels:
-			attrs = Set(iter(r.getAttributes()))
+			attrs = set(iter(r.getAttributes()))
 			if required.issubset(attrs):
 				tmp.append(r)
 		return tmp
@@ -251,6 +255,20 @@ class Entity(object):
 		else:
 			self._tags[tag.value] = tag
 
+	def getRating(self):
+		"""Return the rating of this Entity.
+		0 = Unrated
+		1 - 5 = Rating
+
+		@return: rating
+		"""
+		return self._rating
+
+	rating = property(getRating, doc='The rating for this entity.')
+
+	def setRating(self, value):
+		self._rating = value
+		
 
 class Artist(Entity):
 	"""Represents an artist.
@@ -285,6 +303,9 @@ class Artist(Entity):
 		self._releases = [ ]
 		self._releasesCount = None
 		self._releasesOffset = None
+		self._releaseGroups = [ ]
+		self._releaseGroupsCount = None
+		self._releaseGroupsOffset = None
 
 	def getType(self):
 		"""Returns the artist's type.
@@ -528,6 +549,145 @@ class Artist(Entity):
 	releasesCount = property(getReleasesCount, setReleasesCount,
 		doc='The total number of releases')
 
+	def getReleaseGroups(self):
+		"""Returns a list of release groups from this artist.
+		
+		@return: a list of L{ReleaseGroup} objects
+		"""
+		return self._releaseGroups
+	
+	releaseGroups = property(getReleaseGroups, doc='The list of release groups')
+	
+	def addReleaseGroup(self, releaseGroup):
+		"""Adds a release group to this artist's list of release groups.
+		
+		@param releaseGroup: a L{ReleaseGroup} object
+		"""
+		self._releaseGroups.append(releaseGroup)
+
+	def getReleaseGroupsOffset(self):
+		"""Returns the offset of the release group list.
+
+		This is used if the release group list is incomplete (ie. the
+		web service only returned part of the result for this artist).
+		Note that the offset value is zero-based, which means release
+		group C{0} is the first release group.
+
+		@return: an integer containing the offset, or None
+
+		@see: L{getReleaseGroups}, L{getReleaseGroupsCount}
+		"""
+		return self._releaseGroupsOffset
+
+	def setReleaseGroupsOffset(self, offset):
+		"""Sets the offset of the release group list.
+
+		@param offset: an integer containing the offset, or None
+
+		@see: L{getReleaseGroupsOffset}
+		"""
+		self._releaseGroupsOffset = offset
+
+	releaseGroupsOffset = property(getReleaseGroupsOffset, setReleaseGroupsOffset,
+		doc='The offset of the release group list.')
+
+	def getReleaseGroupsCount(self):
+		"""Returns the number of existing release groups.
+
+		This may or may not match with the number of elements that
+		L{getReleaseGroups} returns. If the count is higher than
+		the list, it indicates that the list is incomplete.
+
+		@return: an integer containing the count, or None
+
+		@see: L{setReleaseGroupsCount}, L{getReleaseGroupsOffset}
+		"""
+		return self._releaseGroupsCount
+
+	def setReleaseGroupsCount(self, value):
+		"""Sets the number of existing release groups.
+
+		@param value: an integer containing the count, or None
+
+		@see: L{getReleaseGroupsCount}, L{setReleaseGroupsOffset}
+		"""
+		self._releaseGroupsCount = value
+
+	releasesCount = property(getReleaseGroupsCount, setReleaseGroupsCount,
+		doc='The total number of release groups')
+
+
+class Rating(object):
+	"""The representation of a MusicBrain rating.
+
+	The rating can have the following values:
+
+	0 = Unrated
+	[1..5] = Rating
+	"""
+	def __init__(self, value=None, count=None):
+		"""Constructor.
+
+		@param value: a string containing the tag's value
+		@param count: the number of users who added this tag
+		"""
+		self._value = value
+		self._count = count
+
+	def getValue(self):
+		"""Returns a string with the tag's value.
+
+		@return: an integer containing the rating's value, or None
+		"""
+		return self._value
+
+	def setValue(self, value):
+		""" Set the value of this rating.
+
+		0 or None = Clear your rating
+		1 - 5 = Rating
+
+		@param value: the rating to apply
+
+		@raise ValueError: if value is not a double or not in the
+		range 0 - 5 or None.
+		"""
+		if value == None:
+		  value = 0
+		try:
+		  value = float(value)
+		except ValueError, e:
+		  raise ValueError("Value for rating needs to be an" \
+		      "float.")
+		if value < 0.0 or value > 5.0:
+		  raise ValueError("Value needs to be in the range [0..5]")
+		self._value = value
+
+	value = property(getValue, setValue, doc='The value of the rating.')
+
+	def getCount(self):
+		"""Returns an integer containing the rating's frequency count.
+
+		@return: an integer containing the rating's frequency count,
+		or None
+		"""
+		return self._count
+	
+	def setCount(self, count):
+		"""Sets the frequency count of this rating.
+		
+		@param count: an integer containing the tag's frequency count
+		"""
+		self._count = count
+		
+	count = property(getCount, setCount, doc="This tag's frequency count.")
+
+	def __str__(self):
+		return str(self._value)
+
+	def __unicode__(self):
+		return unicode(self._value)
+
 
 class Tag(object):
 	"""The representation of a MusicBrainz folksonomy tag.
@@ -823,8 +983,8 @@ class Release(Entity):
 	@note: The current MusicBrainz server implementation supports only a
 	limited set of types.
 	"""
-	# TODO: we need a type for NATs
 	TYPE_NONE = NS_MMD_1 + 'None'
+	TYPE_NON_ALBUM_TRACKS = NS_MMD_1 + "NonAlbum Track"
 
 	TYPE_ALBUM = NS_MMD_1 + 'Album'
 	TYPE_SINGLE = NS_MMD_1 + 'Single'
@@ -858,6 +1018,7 @@ class Release(Entity):
 		self._artist = None
 		self._releaseEvents = [ ]
 		#self._releaseEventsCount = None
+		self._releaseGroup = None
 		self._discs = [ ]
 		#self._discIdsCount = None
 		self._tracks = [ ]
@@ -997,6 +1158,23 @@ class Release(Entity):
 	artist = property(getArtist, setArtist,
 		doc='The main artist of this release.')
 
+	def getReleaseGroup(self):
+		"""Returns the release group to which this release belongs.
+		
+		@return: a L{ReleaseGroup} object, or None.
+		"""
+		return self._releaseGroup
+
+	def setReleaseGroup(self, releaseGroup):
+		"""Sets the release's release group.
+		
+		@param releaseGroup: a L{ReleaseGroup} object, or None.
+		"""
+		self._releaseGroup = releaseGroup
+
+	releaseGroup = property(getReleaseGroup, setReleaseGroup,
+		doc='The release group this release belongs to.')
+
 	def isSingleArtistRelease(self):
 		"""Checks if this is a single artist's release.
 
@@ -1216,6 +1394,158 @@ class Release(Entity):
 	#	self._discIdsCount = value
 
 
+class ReleaseGroup(Entity):
+	"""Represents a ReleaseGroup.
+
+	A ReleaseGroup in MusicBrainz is an L{Entity} which groups several different
+	versions of L{Release} objects (e.g., different editions of the same album).
+
+	@see: L{Release}
+	@see: L{Entity}
+	"""
+
+	def __init__(self, id_=None, title=None):
+		"""Constructor.
+
+		@param id_: a string containing an absolute URI
+		@param title: a string containing the title
+		"""
+		Entity.__init__(self, id_)
+		self._title = title
+		self._id = id_
+		self._type = None
+		self._releases = [ ]
+		self._artist = None
+		self._releasesOffset = 0
+		self._releasesCount = 0
+
+	def getType(self):
+		"""Returns the type of this release group.
+
+		To test for release types, you can use the constants
+		L{Release.TYPE_ALBUM}, L{Release.TYPE_SINGLE}, etc.
+
+		@return: a string containing an absolute URI, or None.
+
+		@see: L{musicbrainz2.utils.getReleaseTypeName}
+		"""
+		return self._type
+
+	def setType(self, type_):
+		"""Sets the type of this release group.
+
+		Use a constant from the L{Release} class, such as
+		L{Release.TYPE_ALBUM} or L{Release.TYPE_SINGLE} to
+		set the value.
+
+		@param type_: a string containing an absolute URI, or None.
+		
+		@see: L{musicbrainz2.utils.getReleaseTypeName}
+		"""
+		self._type = type_
+
+	type = property(getType, setType,
+		doc = 'The type of this release group.')
+
+	def getReleases(self):
+		"""Gets the releases in this release group.
+
+		@return: a list of L{Release} objects
+		@see: L{Release}
+		"""
+		return self._releases
+
+	releases = property(getReleases,
+		doc = 'The list of releases in this release group.')
+
+	def addRelease(self, release):
+		"""Adds a L{Release} to this release group.
+
+		@param release: a L{Release} object
+		"""
+		self._releases.append(release)
+
+	def getReleasesOffset(self):
+		"""Returns the offset of the release list.
+
+		This is used if the release list is incomplete (i.e., the web
+		service only returned a portion of the releases in this release
+		group).
+
+		@return: an integer containing the offset, or None.
+		@see: L{getReleases}, L{getReleasesCount}
+		"""
+		return self._releasesOffset
+
+	def setReleasesOffset(self, offset):
+		"""Sets the offset of the release list.
+
+		@param offset: an integer containing the offset, or None.
+		@see: L{getReleases}, L{getReleasesOffset}
+		"""
+		self._releasesOffset = offset
+
+	releasesOffset = property(getReleasesOffset, setReleasesOffset,
+		doc='The offset of the release list.')
+
+	def getReleasesCount(self):
+		"""Returns the number of releases in this release group.
+
+		This may or may not match the number of elements returned by
+		L{getReleases}.  If the count is higher than the length of that
+		list, then the list is incomplete.
+
+		@return: an integer containing the count, or None
+		@see: L{getReleases}, L{setReleasesCount}, L{getReleasesOffset}
+		"""
+		return self._releasesCount
+
+	def setReleasesCount(self, value):
+		"""Sets the number of releases in this release group.
+
+		@param value: an integer containing the count, or None.
+		@see: L{getReleases}, L{getReleasesCount}, L{getReleasesOffset}
+		"""
+		self._releasesCount = value
+
+	releasesCount = property(getReleasesCount, setReleasesCount,
+		doc = 'The total number of releases')
+
+	def getTitle(self):
+		"""Returns this release group's title.
+
+		@return: a string containing the release group's title
+		"""
+		return self._title
+
+	def setTitle(self, title):
+		"""Sets the release group's title.
+
+		@param title: a string containing the release group's title.
+		"""
+		self._title = title
+
+	title = property(getTitle, setTitle,
+		doc = 'The title of this release group.')
+
+	def getArtist(self):
+		"""Returns the main artist of this release group.
+
+		@return: an L{Artist} object, or None
+		"""
+		return self._artist
+
+	def setArtist(self, artist):
+		"""Sets the release group's main artist.
+
+		@param artist: an L{Artist} object
+		"""
+		self._artist = artist
+
+	artist = property(getArtist, setArtist,
+		doc = 'The main artist of this release group')
+
+
 class Track(Entity):
 	"""Represents a track.
 
@@ -1243,6 +1573,7 @@ class Track(Entity):
 		self._duration = None
 		self._puids = [ ]
 		self._releases = [ ]
+		self._isrcs = [ ]
 
 	def getTitle(self):
 		"""Returns the track's title.
@@ -1335,6 +1666,22 @@ class Track(Entity):
 		@param puid: a string containing a PUID
 		"""
 		self._puids.append(puid)
+	
+	def getISRCs(self):
+		"""Returns the ISRCs associated with this track.
+
+		@return: a list of strings, each containing one ISRC
+		"""
+		return self._isrcs
+
+	isrcs = property(getISRCs, doc='The list of associated ISRCs')
+
+	def addISRC(self, isrc):
+		"""Add a ISRC to this track.
+
+		@param isrc: a string containing an ISRC
+		"""
+		self._isrcs.append(isrc)
 
 	def getReleases(self):
 		"""Returns the list of releases this track appears on.
@@ -1745,6 +2092,109 @@ class ReleaseEvent(object):
 		doc='The format of the release medium.')
 
 
+class CDStub(object):
+	"""Represents a CD Stub"""
+
+	def __init__(self, disc):
+		"""Constructor.
+
+		@param disc: a L{Disc} object to create this CD Stub from
+		"""
+		assert isinstance(disc, Disc), 'musicbrainz2.model.Disc expected'
+		self._disc = disc
+		self._tracks = [ ]
+		self._title = ""
+		self._artist = ""
+		self._barcode = ""
+		self._comment = ""
+
+	def setTitle(self, title):
+		"""Sets the title of this release.
+
+		@param title: a string containing the title
+		"""
+		self._title = title
+
+	def getTitle(self):
+		"""Returns the title of this release.
+
+		@return: a string containing the title
+		"""
+		return self._title
+
+	title = property(getTitle, setTitle,
+	       doc='The title of the release')
+
+	def setArtist(self, artist):
+		"""Sets the artist of this release.
+
+		@param artist: a string containing the artist
+		"""
+		self._artist = artist
+
+	def getArtist(self):
+		"""Returns the artist of this release.
+
+		@return: a string containing the artist
+		"""
+		return self._artist
+
+	artist = property(getArtist, setArtist,
+		doc='The artist of the release')
+
+	def setComment(self, comment):
+		"""Sets the comment for this release.
+
+		@param comment: a string containing the comment
+		"""
+		self._comment = comment
+
+	def getComment(self):
+		"""Returns the comment for this release.
+
+		@return: a string containing the comment
+		"""
+		return self._comment
+
+	comment = property(getComment, setComment,
+		 doc='Comment for the release (optional)')
+
+	def setBarcode(self, barcode):
+		"""Sets the barcode of this release.
+
+		@param barcode: a string containing the barcode
+		"""
+		self._barcode = barcode
+
+	def getBarcode(self):
+		"""Returns the barcode of this release.
+
+		@return: a string containing the barcode
+		"""
+		return self._barcode
+
+	barcode = property(getBarcode, setBarcode,
+		 doc='Barcode for the release (optional)')
+
+	def addTrack(self, title, artist=''):
+		"""Add a track to this release
+
+		@param title: a string containing the title of the track
+		@param artist: a string containing the artist of the track, 
+		               if different to the album artist
+		"""
+		self._tracks.append((title, artist))
+
+	def getTracks(self):
+		"""Return all the tracks on the release.
+
+		@return: a list of tuples containing (title, artist) pairs
+		         for each track
+		"""
+		return self._tracks
+
+	tracks = property(getTracks, doc='The tracks of the release.')
+
 class Disc(object):
 	"""Represents an Audio CD.
 
