@@ -12,7 +12,7 @@ model <musicbrainz2.model>}.
 
 @author: Matthias Friedrich <matt@mafr.de>
 """
-__revision__ = '$Id: webservice.py 9715 2008-03-01 09:21:28Z matt $'
+__revision__ = '$Id: webservice.py 12708 2010-03-15 17:45:25Z matt $'
 
 import re
 import urllib
@@ -30,9 +30,9 @@ __all__ = [
 	'WebServiceError', 'AuthenticationError', 'ConnectionError',
 	'RequestError', 'ResourceNotFoundError', 'ResponseError', 
 	'IIncludes', 'ArtistIncludes', 'ReleaseIncludes', 'TrackIncludes',
-	'LabelIncludes',
+	'LabelIncludes', 'ReleaseGroupIncludes',
 	'IFilter', 'ArtistFilter', 'ReleaseFilter', 'TrackFilter',
-	'UserFilter', 'LabelFilter',
+	'UserFilter', 'LabelFilter', 'ReleaseGroupFilter',
 	'IWebService', 'WebService', 'Query',
 ]
 
@@ -412,13 +412,71 @@ class LabelFilter(IFilter):
 	def createParameters(self):
 		return _createParameters(self._params)
 
+class ReleaseGroupFilter(IFilter):
+	"""A filter for the release group collection."""
+
+	def __init__(self, title=None, releaseTypes=None, artistName=None,
+			artistId=None, limit=None, offset=None, query=None):
+		"""Constructor.
+
+		If C{artistId} is set, only releases matching those IDs are
+		returned.  The C{releaseTypes} parameter allows you to limit
+		the types of the release groups returned. You can set it to
+		C{(Release.TYPE_ALBUM, Release.TYPE_OFFICIAL)}, for example,
+		to only get officially released albums. Note that those values
+		are connected using the I{AND} operator. MusicBrainz' support
+		is currently very limited, so C{Release.TYPE_LIVE} and
+		C{Release.TYPE_COMPILATION} exclude each other (see U{the
+		documentation on release attributes
+		<http://wiki.musicbrainz.org/AlbumAttribute>} for more
+		information and all valid values).
+
+		If both the C{artistName} and the C{artistId} parameter are
+		given, the server will ignore C{artistName}.
+
+		The C{query} parameter may contain a query in U{Lucene syntax
+		<http://lucene.apache.org/java/docs/queryparsersyntax.html>}.
+		Note that C{query} may not be used together with the other
+		parameters except for C{limit} and C{offset}.
+
+		@param title: a unicode string containing the release group's title
+		@param releaseTypes: a sequence of release type URIs
+		@param artistName: a unicode string containing the artist's name
+		@param artistId: a unicode string containing the artist's ID
+		@param limit: the maximum number of release groups to return
+		@param offset: start results at this zero-based offset
+		@param query: a string containing a query in Lucene syntax
+
+		@see: the constants in L{musicbrainz2.model.Release}
+		"""
+		if releaseTypes is None or len(releaseTypes) == 0:
+			releaseTypesStr = None
+		else:
+			releaseTypesStr = ' '.join(map(mbutils.extractFragment, releaseTypes))
+
+		self._params = [
+			('title', title),
+			('releasetypes', releaseTypesStr),
+			('artist', artistName),
+			('artistid', mbutils.extractUuid(artistId)),
+			('limit', limit),
+			('offset', offset),
+			('query', query),
+		]
+
+		if not _paramsValid(self._params):
+			raise ValueError('invalid combination of parameters')
+
+	def createParameters(self):
+		return _createParameters(self._params)
+
 
 class ReleaseFilter(IFilter):
 	"""A filter for the release collection."""
 
 	def __init__(self, title=None, discId=None, releaseTypes=None,
 			artistName=None, artistId=None, limit=None,
-			offset=None, query=None):
+			offset=None, query=None, trackCount=None):
 		"""Constructor.
 
 		If C{discId} or C{artistId} are set, only releases matching
@@ -449,6 +507,7 @@ class ReleaseFilter(IFilter):
 		@param limit: the maximum number of releases to return
 		@param offset: start results at this zero-based offset
 		@param query: a string containing a query in Lucene syntax
+		@param trackCount: the number of tracks in the release
 
 		@see: the constants in L{musicbrainz2.model.Release}
 		"""
@@ -463,10 +522,11 @@ class ReleaseFilter(IFilter):
 			('discid', discId),
 			('releasetypes', releaseTypesStr),
 			('artist', artistName),
-			('artistid', artistId),
+			('artistid', mbutils.extractUuid(artistId)),
 			('limit', limit),
 			('offset', offset),
 			('query', query),
+			('count', trackCount),
 		]
 
 		if not _paramsValid(self._params):
@@ -510,9 +570,9 @@ class TrackFilter(IFilter):
 		self._params = [
 			('title', title),
 			('artist', artistName),
-			('artistid', artistId),
+			('artistid', mbutils.extractUuid(artistId)),
 			('release', releaseTitle),
-			('releaseid', releaseId),
+			('releaseid', mbutils.extractUuid(releaseId)),
 			('duration', duration),
 			('puid', puid),
 			('limit', limit),
@@ -571,7 +631,8 @@ class ArtistIncludes(IIncludes):
 	"""
 	def __init__(self, aliases=False, releases=(), vaReleases=(),
 			artistRelations=False, releaseRelations=False,
-			trackRelations=False, urlRelations=False, tags=False):
+			trackRelations=False, urlRelations=False, tags=False,
+			ratings=False, releaseGroups=False):
 
 		assert not isinstance(releases, basestring)
 		assert not isinstance(vaReleases, basestring)
@@ -580,10 +641,12 @@ class ArtistIncludes(IIncludes):
 		self._includes = {
 			'aliases':		aliases,
 			'artist-rels':		artistRelations,
+			'release-groups':	releaseGroups,
 			'release-rels':		releaseRelations,
 			'track-rels':		trackRelations,
 			'url-rels':		urlRelations,
-			'tags':		tags,
+			'tags':			tags,
+			'ratings':		ratings,
 		}
 
 		for elem in releases:
@@ -602,11 +665,13 @@ class ReleaseIncludes(IIncludes):
 			discs=False, tracks=False,
 			artistRelations=False, releaseRelations=False,
 			trackRelations=False, urlRelations=False,
-			labels=False, tags=False):
+			labels=False, tags=False, ratings=False, isrcs=False,
+			releaseGroup=False):
 		self._includes = {
 			'artist':		artist,
 			'counts':		counts,
 			'labels':		labels,
+			'release-groups':	releaseGroup,
 			'release-events':	releaseEvents,
 			'discs':		discs,
 			'tracks':		tracks,
@@ -614,13 +679,36 @@ class ReleaseIncludes(IIncludes):
 			'release-rels':		releaseRelations,
 			'track-rels':		trackRelations,
 			'url-rels':		urlRelations,
-			'tags':		tags,
+			'tags':			tags,
+			'ratings':		ratings,
+			'isrcs':		isrcs,
 		}
 
 		# Requesting labels without releaseEvents makes no sense,
 		# so we pull in releaseEvents, if necessary.
 		if labels and not releaseEvents:
 			self._includes['release-events'] = True
+		# Ditto for isrcs with no tracks
+		if isrcs and not tracks:
+			self._includes['tracks'] = True
+
+	def createIncludeTags(self):
+		return _createIncludes(self._includes)
+
+
+class ReleaseGroupIncludes(IIncludes):
+	"""A specification on how much data to return with a release group."""
+
+	def __init__(self, artist=False, releases=False, tags=False):
+		"""Constructor.
+
+		@param artist: Whether to include the release group's main artist info.
+		@param releases: Whether to include the release group's releases.
+		"""
+		self._includes = {
+			'artist':		artist,
+			'releases':		releases,
+		}
 
 	def createIncludeTags(self):
 		return _createIncludes(self._includes)
@@ -630,7 +718,8 @@ class TrackIncludes(IIncludes):
 	"""A specification on how much data to return with a track."""
 	def __init__(self, artist=False, releases=False, puids=False,
 			artistRelations=False, releaseRelations=False,
-			trackRelations=False, urlRelations=False, tags=False):
+			trackRelations=False, urlRelations=False, tags=False,
+			ratings=False, isrcs=False):
 		self._includes = {
 			'artist':		artist,
 			'releases':		releases,
@@ -639,7 +728,9 @@ class TrackIncludes(IIncludes):
 			'release-rels':		releaseRelations,
 			'track-rels':		trackRelations,
 			'url-rels':		urlRelations,
-			'tags':		tags,
+			'tags':			tags,
+			'ratings':		ratings,
+			'isrcs':		isrcs,
 		}
 
 	def createIncludeTags(self):
@@ -648,10 +739,11 @@ class TrackIncludes(IIncludes):
 
 class LabelIncludes(IIncludes):
 	"""A specification on how much data to return with a label."""
-	def __init__(self, aliases=False, tags=False):
+	def __init__(self, aliases=False, tags=False, ratings=False):
 		self._includes = {
 			'aliases':		aliases,
-			'tags':		tags,
+			'tags':			tags,
+			'ratings':		ratings,
 		}
 
 	def createIncludeTags(self):
@@ -948,13 +1040,51 @@ class Query(object):
 		"""
 		result = self._getFromWebService('release', '', filter=filter)
 		return result.getReleaseResults()
+	
+	def getReleaseGroupById(self, id_, include=None):
+		"""Returns a release group.
 
+		If no release group with that ID can be found, C{include}
+		contains invalid tags, or there's a server problem, an
+		exception is raised.
+
+		@param id_: a string containing the release group's ID
+		@param include: a L{ReleaseGroupIncludes} object, or None
+
+		@return: a L{ReleaseGroup <musicbrainz2.model.ReleaseGroup>} object, or None
+
+		@raise ConnectionError: couldn't connect to server
+		@raise RequestError: invalid ID or include tags
+		@raise ResourceNotFoundError: release doesn't exist
+		@raise ResponseError: server returned invalid data
+		"""
+		uuid = mbutils.extractUuid(id_, 'release-group')
+		result = self._getFromWebService('release-group', uuid, include)
+		releaseGroup = result.getReleaseGroup()
+		if releaseGroup is not None:
+			return releaseGroup
+		else:
+			raise ResponseError("server didn't return releaseGroup")
+
+	def getReleaseGroups(self, filter):
+		"""Returns release groups matching the given criteria.
+		
+		@param filter: a L{ReleaseGroupFilter} object
+		
+		@return: a list of L{musicbrainz2.wsxml.ReleaseGroupResult} objects
+		
+		@raise ConnectionError: couldn't connect to server
+		@raise RequestError: invalid ID or include tags
+		@raise ResponseError: server returned invalid data
+		"""
+		result = self._getFromWebService('release-group', '', filter=filter)
+		return result.getReleaseGroupResults()
 
 	def getTrackById(self, id_, include=None):
 		"""Returns a track.
 
 		If no track with that ID can be found, C{include} contains
-		invalid tags or there's a server problem, and exception is
+		invalid tags or there's a server problem, an exception is
 		raised.
 
 		@param id_: a string containing the track's ID
@@ -1054,7 +1184,7 @@ class Query(object):
 		@param tracks2puids: a dictionary mapping track IDs to PUIDs
 
 		@raise ConnectionError: couldn't connect to server
-		@raise RequestError: invalid track- or PUIDs
+		@raise RequestError: invalid track or PUIDs
 		@raise AuthenticationError: invalid user name and/or password
 		"""
 		assert self._clientId is not None, 'Please supply a client ID'
@@ -1068,7 +1198,107 @@ class Query(object):
 		encodedStr = urllib.urlencode(params, True)
 
 		self._ws.post('track', '', encodedStr)
+	
+	def submitISRCs(self, tracks2isrcs):
+		"""Submit track to ISRC mappings.
 
+		The C{tracks2isrcs} parameter has to be a dictionary, with the
+		keys being MusicBrainz track IDs (either as absolute URIs or
+		in their 36 character ASCII representation) and the values
+		being ISRCs (ASCII, 12 characters).
+
+		Note that this method only works if a valid user name and
+		password have been set. See the example in L{Query} on how
+		to supply authentication data.
+
+		@param tracks2isrcs: a dictionary mapping track IDs to ISRCs
+
+		@raise ConnectionError: couldn't connect to server
+		@raise RequestError: invalid track or ISRCs
+		@raise AuthenticationError: invalid user name and/or password
+		"""
+		params = [ ]
+
+		for (trackId, isrc) in tracks2isrcs.iteritems():
+			trackId = mbutils.extractUuid(trackId, 'track')
+			params.append( ('isrc', trackId + ' ' + isrc) )
+
+		encodedStr = urllib.urlencode(params, True)
+
+		self._ws.post('track', '', encodedStr)
+
+	def addToUserCollection(self, releases):
+		"""Add releases to a user's collection.
+
+		The releases parameter must be a list. It can contain either L{Release}
+		objects or a string representing a MusicBrainz release ID (either as
+		absolute URIs or in their 36 character ASCII representation).
+
+		Adding a release that is already in the collection has no effect.
+
+		@param releases: a list of releases to add to the user collection
+
+		@raise ConnectionError: couldn't connect to server
+		@raise AuthenticationError: invalid user name and/or password
+		"""
+		rels = []
+		for release in releases:
+			if isinstance(release, Release):
+				rels.append(mbutils.extractUuid(release.id))
+			else:
+				rels.append(mbutils.extractUuid(release))
+		encodedStr = urllib.urlencode({'add': ",".join(rels)}, True)
+		self._ws.post('collection', '', encodedStr)
+
+	def removeFromUserCollection(self, releases):
+		"""Remove releases from a user's collection.
+
+		The releases parameter must be a list. It can contain either L{Release}
+		objects or a string representing a MusicBrainz release ID (either as
+		absolute URIs or in their 36 character ASCII representation).
+
+		Removing a release that is not in the collection has no effect.
+
+		@param releases: a list of releases to remove from the user collection
+
+		@raise ConnectionError: couldn't connect to server
+		@raise AuthenticationError: invalid user name and/or password
+		"""
+		rels = []
+		for release in releases:
+			if isinstance(release, Release):
+				rels.append(mbutils.extractUuid(release.id))
+			else:
+				rels.append(mbutils.extractUuid(release))
+		encodedStr = urllib.urlencode({'remove': ",".join(rels)}, True)
+		self._ws.post('collection', '', encodedStr)
+
+	def getUserCollection(self, offset=0, maxitems=100):
+		"""Get the releases that are in a user's collection
+		
+		A maximum of 100 items will be returned for any one call
+		to this method. To fetch more than 100 items, use the offset
+		parameter.
+
+		@param offset: the offset to start fetching results from
+		@param maxitems: the upper limit on items to return
+
+		@return: a list of L{musicbrainz2.wsxml.ReleaseResult} objects
+
+		@raise ConnectionError: couldn't connect to server
+		@raise AuthenticationError: invalid user name and/or password
+		"""
+		params = { 'offset': offset, 'maxitems': maxitems }
+		
+		stream = self._ws.get('collection', '', filter=params)
+		print stream
+		try:
+			parser = MbXmlParser()
+			result = parser.parse(stream)
+		except ParseError, e:
+			raise ResponseError(str(e), e)
+		
+		return result.getReleaseResults()
 
 	def submitUserTags(self, entityUri, tags):
 		"""Submit folksonomy tags for an entity.
@@ -1133,7 +1363,112 @@ class Query(object):
 			raise ResponseError(str(e), e)
 		
 		return result.getTagList()
+
+	def submitUserRating(self, entityUri, rating):
+		"""Submit rating for an entity.
+
+		Note that all previously existing rating from the authenticated
+		user are replaced with the one given to this method. Other
+		users' ratings are not affected.
 		
+		@param entityUri: a string containing an absolute MB ID
+		@param rating: A L{Rating <musicbrainz2.model.Rating>} object
+		             or integer
+
+		@raise ValueError: invalid entityUri
+		@raise ConnectionError: couldn't connect to server
+		@raise RequestError: invalid ID, entity or tags
+		@raise AuthenticationError: invalid user name and/or password
+		"""
+		entity = mbutils.extractEntityType(entityUri)
+		uuid = mbutils.extractUuid(entityUri, entity)
+		params = (
+			('type', 'xml'),
+			('entity', entity),
+			('id', uuid),
+			('rating', unicode(rating).encode('utf-8'))
+		)
+
+		encodedStr = urllib.urlencode(params)
+
+		self._ws.post('rating', '', encodedStr)
+
+
+	def getUserRating(self, entityUri):
+		"""Return the rating a user has applied to an entity.
+
+		The given parameter has to be a fully qualified MusicBrainz
+		ID, as returned by other library functions.
+		
+		Note that this method only works if a valid user name and
+		password have been set. Only the rating the authenticated user
+		applied to the entity will be returned. If username and/or
+		password are incorrect, an AuthenticationError is raised.
+		
+		This method will return a L{Rating <musicbrainz2.model.Rating>}
+		object.
+		
+		@param entityUri: a string containing an absolute MB ID
+		
+		@raise ValueError: invalid entityUri
+  		@raise ConnectionError: couldn't connect to server
+		@raise RequestError: invalid ID or entity
+		@raise AuthenticationError: invalid user name and/or password
+		"""
+		entity = mbutils.extractEntityType(entityUri)
+		uuid = mbutils.extractUuid(entityUri, entity)
+		params = { 'entity': entity, 'id': uuid }
+		
+		stream = self._ws.get('rating', '', filter=params)
+		try:
+			parser = MbXmlParser()
+			result = parser.parse(stream)
+		except ParseError, e:
+			raise ResponseError(str(e), e)
+		
+		return result.getRating()
+
+	def submitCDStub(self, cdstub):
+		"""Submit a CD Stub to the database.
+
+		The number of tracks added to the CD Stub must match the TOC and DiscID
+		otherwise the submission wil fail. The submission will also fail if 
+		the Disc ID is already in the MusicBrainz database.
+
+		This method will only work if no user name and password are set.
+
+		@param cdstub: a L{CDStub} object to submit
+		
+		@raise RequestError: Missmatching TOC/Track information or the
+		       the CD Stub already exists or the Disc ID already exists
+		"""
+		assert self._clientId is not None, 'Please supply a client ID'
+		disc = cdstub._disc
+		params = [ ]
+		params.append( ('client', self._clientId.encode('utf-8')) )
+		params.append( ('discid', disc.id) )
+		params.append( ('title', cdstub.title) )
+		params.append( ('artist', cdstub.artist) )
+		if cdstub.barcode != "":
+			params.append( ('barcode', cdstub.barcode) )
+		if cdstub.comment != "":
+			params.append( ('comment', cdstub.comment) )
+
+		trackind = 0
+		for track,artist in cdstub.tracks:
+			params.append( ('track%d' % trackind, track) )
+			if artist != "":
+				params.append( ('artist%d' % trackind, artist) )
+
+			trackind += 1
+
+		toc = "%d %d %d " % (disc.firstTrackNum, disc.lastTrackNum, disc.sectors)
+	        toc = toc + ' '.join( map(lambda x: str(x[0]), disc.getTracks()) )
+
+		params.append( ('toc', toc) )
+
+		encodedStr = urllib.urlencode(params)
+		self._ws.post('release', '', encodedStr)
 
 def _createIncludes(tagMap):
 	selected = filter(lambda x: x[1] == True, tagMap.items())
