@@ -19,5 +19,18 @@ class ReleaseFilterTest(unittest.TestCase):
 		self.assert_( ('artist', 'Tori Amos') )
 		self.assert_( ('releasetypes', 'Album Official') )
 
+	def testQuery(self):
+		try:
+			f1 = ReleaseFilter(title='Pink', query='xyz')
+			self.fail('title and query are mutually exclusive')
+		except ValueError:
+			pass
+		except:
+			self.fail('invalid exception')
+
+		# the following shouldn't throw exceptions:
+		f2 = ReleaseFilter(title='Pink', limit=10, offset=20)
+		f3 = ReleaseFilter(query='Pink', limit=10, offset=20)
+
 
 # EOF
